
//
//
// Praktikum 3 in Algorithmen und Datenstrukturen / Fr. Prof. Dr. Kirch-Prinz
// Binrer Baum
//
// made by
//
// Rene Valdez
// Peter Ellenberg
//
//
// BinBaum.h Headerdatei der Klassen Cell<T>, Node<T>, Heap<T>, BadAtom, NotFound<T>
// 

#include <string>
#include <iostream>
#include <iomanip>
#include <cmath>

// Nullwert
#define NULL 0

// Pseudowert
#define PSEUDO -1

using namespace std;

template<class T>
class Cell {

	private:
		Cell<T>* leftChild;
		Cell<T>* rightChild;
	public:
		// Konstruktor
		Cell(Cell<T>* leftChild, Cell<T>* rightChild) { this->leftChild = leftChild; this->rightChild = rightChild; };
		// Konstruktor
		Cell(void) { leftChild = NULL; rightChild = NULL; };
		// gibt den linken Kindknoten zurck
		Cell<T>* getLeftChild() { return leftChild; };
		// gibt den rechten Kindknoten zurck
		Cell<T>* getRightChild() {return rightChild; };
		// setzt den linken Kindknoten
		void setLeftChild(Cell<T>* cell) { leftChild = cell; };
		// setzt den rechten Kindknoten
		void setRightChild(Cell<T>* cell) { rightChild = cell; };

		virtual T getAtom(void) = 0;
		virtual void setAtom(T atom) = 0;
};

template<class T>
class Node : public Cell<T> {
	private:
		T atom;

	public:
		// Konstruktor
		Node(T atom) { 	this->atom = atom; };
		// gibt ein Atom zurck
		T getAtom(void) {return atom; };
		// setzt ein neues Atom
		void setAtom(T atom) {this->atom = atom; };
};


template<class T>
class Heap {
	private:
		Cell<T>* wurzel;
		long anzahl;

		// setzt die Anzahl der Knoten im Baum
		void setAnzahl(long anzahl) { this->anzahl = anzahl; };
		// ---------------------------------------------------------------------------
		// gibt die Anzahl der Knoten im Baum zurck
		long getAnzahl(void) { return anzahl; };
		// ---------------------------------------------------------------------------
		// addiert zur Anzahl der Knoten im Baum eins dazu
		void addAnzahl() {anzahl++;};
		// ---------------------------------------------------------------------------
		// subtrahiert die Anzahl der Knoten im Baum um eins
		void subAnzahl() {anzahl--;};
		// ---------------------------------------------------------------------------
		// lscht einen Knoten
		void deleteNode(Cell* node);
		// ---------------------------------------------------------------------------
		// gibt den Vaterknoten fr die Identifikationsnummer eines Knotens im Baum zurck
		Cell<T>* getParent(long ident) {
			
			Cell<T>* parent;
			
			// wir sind an der Wurzel angekommen
			if (ident == 1 ||ident == 0) {
				parent = wurzel;
				return parent;
			} else {
				// wir suchen nach dem nchsten Vaterknoten
				parent = getParent((long)ident/2);
			}

			// wir gehen den Weg wieder zurck, und hangeln uns durch den Baum,
			// solange es Kindknoten gibt.
			if (ident%2 == 0) {
				if (parent->getLeftChild() != 0) {
					parent = parent->getLeftChild();
				}
			} else if (ident%2 == 1) {
				if (parent->getRightChild() != 0) {
					parent = parent->getRightChild();
				}
			}
			return parent;
		};
		// ---------------------------------------------------------------------------
		// Verdopplung eines Heaps
		void cloneMe(Cell<T>* nodeNeu, Cell<T>* nodeAlt) {
			nodeNeu->setAtom(nodeAlt->getAtom());
			if (nodeAlt->getLeftChild() != NULL) {
				Node<T>* newLeftNode = new Node<T>(0); 
				nodeNeu->setLeftChild((Cell<T>*)newLeftNode);
				cloneMe((Cell<T>*)newLeftNode, nodeAlt->getLeftChild());
			}
			if (nodeAlt->getRightChild() != NULL) {
				Node<T>* newRightNode = new Node<T>(0);
				nodeNeu->setRightChild((Cell<T>*)newRightNode);
				cloneMe((Cell<T>*)newRightNode, nodeAlt->getRightChild());
			}
		};
		// ---------------------------------------------------------------------------
		// Tauschen von Atomen fr den Bubble-Down
		void swap(Cell<T>* parent, Cell<T>* child) {
			T atom = parent->getAtom(); 
			parent->setAtom(child->getAtom()); 
			child->setAtom(atom); 
		};
		// ---------------------------------------------------------------------------
		// sucht im Baum nach einem Atom und gibt die Identifikationsnummer fr einen Knoten zurck
		long searchAtom(Cell<T>* node, T atom, long ident) {
			Cell<T>* child;
			int found;
			
			// zuerst diverse Abbruchbedingungen fr die Rekursion:

			// Atom wurde gefunden
			if (node->getAtom()==atom)
			{
				return ident;
			}

			// es gibt keinen linken Kindknoten mehr
			if (node->getLeftChild()==NULL) {
				return 0;
			}

			// der Knoten ist grer als das zu suchende Atom 
			// (die darunterliegenden Knoten sind alle grer)
			if (node->getAtom() > atom) {
				return 0;
			}
			
			// wir suchen immer linksseitig
			child = node->getLeftChild();
			
			// sollte hier noch was schiefgehen, gehen wir zurck 
			// (kann eigentlich gar nicht passieren, siehe Abbruchbedingungen)
			if (child == NULL) return ident;
			
			// rekursiver Aufruf
			found = searchAtom(child, atom, ident*2);

			if (found == 0) {
				// es wurde noch nichts gefunden
				child = node->getRightChild();
				if (child == NULL) return found;
				found = searchAtom(child, atom, ident*2+1);

			}
			return found;
		};
		// ---------------------------------------------------------------------------
		// Bubbelt die Atome im Baum von unten nach oben. 
		// D.h. kleinere Werte werden mit Atomen in Elternknoten vertauscht,
		// falls diese grer sind (rekursiv)

		Cell<T>* bubbleUpAtom(long ident, T atom) {
				
			static T newAtom;
			T temp;
			newAtom = atom;
			Cell<T>* parent;

			// wir sind an der Wurzel angekommen
			if (ident == 1) {
				parent = wurzel;
				return parent;

			} else {
				// wir gehen so lange im Baum nach oben, bis wir an der Wurzel ankommen
				parent = bubbleUpAtom((long)ident/2, newAtom);
			}

			// ist das neue Atom kleiner als das Atom im aktuellen Knoten ?
			if (newAtom < parent->getAtom()) {
				// hier wird in dem aktuellen Knoten das kleiner Atom geschrieben,
				// das alte Atom wird sich in newAtom gemerkt
				temp = parent->getAtom();
				parent->setAtom(newAtom);
				newAtom = temp;
			}
			
			// wir gehen den Baum wieder nach unten, muessen aber entscheiden, ob wir nach
			// links oder nach rechts gehen. Dort wird dann das
			if (ident%2 == 0) {
				// hier wird dann endgltig der Tausch der Atome durchgefhrt.
				// der "Bubble" ist fertig
				parent = parent->getLeftChild();
				temp = parent->getAtom();
				parent->setAtom(newAtom);
				newAtom = temp;

			} else if (ident%2 == 1) {
				// hier wird dann endgltig der Tausch der Atome durchgefhrt.
				// der "Bubble" ist fertig
				parent = parent->getRightChild();
				temp = parent->getAtom();
				parent->setAtom(newAtom);
				newAtom = temp;
			}

			return parent;
		};
		// ---------------------------------------------------------------------------
		// das Atom in der Wurzel, das sicher grer ist als die Kinder nach
		// dem Tausch mit dem letzten Atom, wird an die richtige Position "gebubblet"
		void bubbleDownAtom(Cell<T>* node) {

			Cell<T>* linkesKind = node->getLeftChild();
			Cell<T>* rechtesKind = node->getRightChild();

			// es gibt keine Kinder mehr -> listo
			if (linkesKind == NULL) return;

			// wenn es ein rechtes Kind gibt
			if (rechtesKind != NULL) {

				// wenn das linke Kind Atom kleiner ist als das Atom des rechten Kindes
				if (linkesKind->getAtom() < rechtesKind->getAtom()) {
					if (node->getAtom()>linkesKind->getAtom()) {
						swap(node, linkesKind);
						bubbleDownAtom(linkesKind);
					}
				// wenn das rechte Kind Atom kleiner ist als das Atom des linken Kindes
				} else {
					if (node->getAtom() > rechtesKind->getAtom()) {
						swap(node, rechtesKind);
						bubbleDownAtom(rechtesKind);
					}
				}

			// wenn es nur ein linkes Kind gibt
			} else {
				if (node->getAtom() > linkesKind->getAtom()) {
					swap(node, linkesKind);
				}
			}
			return;
		};
		// ---------------------------------------------------------------------------
		// lscht den ganzen Baum
		void delTree(Cell<T>* node) {
			if (node->getLeftChild() != NULL) {
				delTree(node->getLeftChild());
			}

			if (node->getRightChild() != NULL) {
				delTree(node->getRightChild());
			}
			delete node;
		};
		// ---------------------------------------------------------------------------

	public:
		// Standardkonstruktor
		Heap() {
			wurzel = NULL;
			anzahl = NULL;
		};
		// ---------------------------------------------------------------------------
		// Kopierkonstruktor
		Heap(const Heap<T>& sourceHeap) {
			if (sourceHeap.anzahl > 0) {
				Node<T>* newNode = new Node<T>(sourceHeap.wurzel->getAtom());
				this->wurzel = (Cell<T>*)newNode;
				this->anzahl = sourceHeap.anzahl; 
				this->cloneMe(wurzel, sourceHeap.wurzel);
			} 
			else 
			{
				this->anzahl = 0;
			}
			
		};
		// ---------------------------------------------------------------------------
		// berladener Zuweisungsoperator
		Heap<T>& operator = (const Heap<T>& sourceHeap) {
			if (wurzel!=NULL) { delTree(wurzel); }
			Node<T>* newNode = new Node<T>(sourceHeap.wurzel->getAtom());
			this->wurzel = (Cell<T>*)newNode;
			this->anzahl = sourceHeap.anzahl; 
			this->cloneMe(wurzel, sourceHeap.wurzel); 
			return *this;
		};
		// ---------------------------------------------------------------------------
		// gibt die Wurzel zurck
		Cell<T>* getWurzel() { return wurzel; };
		// ---------------------------------------------------------------------------
		// fgt ein neues Atom (und einen neuen Knoten) in den Baum ein
		void insert(T atom) {

			Node<T>* newNode;
			BadAtom badAtom;

			if (anzahl == 0) {
				addAnzahl();
				newNode = new Node<T>(atom);
				this->wurzel = (Cell<T>*)newNode;
				return;
			}
			
			try {
				// sucht, ob das Atom bereits vorhanden ist
				find(atom);

			// falls kein Atom gefunden wurde
			} catch (NotFound<T>) {
				
				// erhht die Anzahl um eins
				addAnzahl();

				// erstellt einen neuen Knoten
				newNode = new Node<T>(atom);

				if (getAnzahl()%2 == 0) {
					// setzt den Pointer im Vaterknoten auf diesen neuen Knoten (linkes Kind)
					getParent( getAnzahl())->setLeftChild((Cell<T>*)newNode );
					// Bubbelt die Atome
					bubbleUpAtom(getAnzahl(), atom);
					return;
				} else {
					// setzt den Pointer im Vaterknoten auf diesen neuen Knoten (rechtes Kind)
					getParent( getAnzahl())->setRightChild((Cell<T>*)newNode );
					// Bubbelt die Atome
					bubbleUpAtom(getAnzahl(), atom);
					return;
				}
			}
			// falls Atom vorhanden, wird eine Exception geworfen
			throw badAtom;
			return;
		};
		// ---------------------------------------------------------------------------
		// lscht die Wurzel
		T deleteMin() {
			Cell<T>* letzterNode;
			Cell<T>* vorletzterNode;
			T kleinstesAtom = 0;

			if (getAnzahl() == 0) {
				NotFound<T> notFound(0);
				throw notFound;
			}
			if (getAnzahl() == 1) {
				kleinstesAtom = wurzel->getAtom();
				delete wurzel;
				subAnzahl();
				return kleinstesAtom;
			} else if (getAnzahl()%2 == 0) {
				vorletzterNode = getParent(getAnzahl()/2);
				letzterNode = vorletzterNode->getLeftChild();
				vorletzterNode->setLeftChild(NULL);
			} else if (getAnzahl()%2 == 1) {
				vorletzterNode = getParent(getAnzahl()/2);
				letzterNode = vorletzterNode->getRightChild();
				vorletzterNode->setRightChild(NULL);
			}
			// vertauschen des Atoms in Wurzel und des letzten Elements
			swap(wurzel, letzterNode);
			kleinstesAtom = letzterNode->getAtom();
			delete letzterNode;
			subAnzahl();
			bubbleDownAtom(wurzel);
			return kleinstesAtom;
		};
		// ---------------------------------------------------------------------------
		// findet ein Atom im Baum
		long find(T atom) {
			NotFound<T> notFound(atom);
			long ident;
			ident = searchAtom(wurzel, atom, 1);
			if (ident == 0) throw notFound;
			return ident;
		};
		// ---------------------------------------------------------------------------
		// gibt den Baum sortiert aus
		void sort() {
			Heap showHeap(*this);
			T* showvector;
			long index = 0;
			long anzahl = showHeap.getAnzahl();

			if (anzahl>0) {
				showvector = new T[anzahl];
				while (showHeap.getAnzahl()!=0) {
					showvector[index]=showHeap.deleteMin();
					index++;
				};
				showVec(showvector, anzahl);
				delete[] showvector;
			} else { cout << "\n" << "Der Baum ist leer.\n\n"; }

		};
		// ---------------------------------------------------------------------------
		// zeigt den Baum unsortiert am Bildschirm an
		void showTree(Heap<T>& tree) {

			T* showvector;
			
			if (tree.getAnzahl()>0) {
				showvector = new T[tree.getAnzahl()];
				runTree(tree.getWurzel(), showvector, 1);
				showVec(showvector, tree.getAnzahl());
			
				delete[] showvector;
			} else { cout << "\n" << "Der Baum ist leer.\n\n"; }
		};

		void showVec(T vec[], long anzahl) {

			string strich = "--------  ----------------------------------------------";
			int ebene = 0;
			long id = 2, i;

			cout << "  ID        " << "Value" << endl;
			cout << strich << endl;
			cout << " " << setw(3) << "1" << "      " << setw(5) << vec[0] << "             " << "Wurzel" << endl;
			cout << strich << endl;

			ebene++;
			i=1;

			for (id; id <= anzahl; id++) {
	
				cout << " " << setw(3) << id << "      " << setw(5) << vec[id-1];
				
				if (i==(pow(2, ebene))) {

					cout << "             Ebene " << ebene << endl;
					cout << strich << endl;
					i=1;
					ebene++;
				}
				else if (id==anzahl) {
					
						cout << "             Ebene " << ebene << "\n\n";

					}
					else {cout << endl; i++;}
				
			};

		};

		//durchluft den Baum und fllt einen Vektor der zur Ausgabe benutzt wird
		void runTree(Cell<T>* wurzel, T vec[], long ident) {
			
			if (ident!=0) {vec[ident-1]=wurzel->getAtom(); }
			if (wurzel->getLeftChild() != NULL) {
				runTree(wurzel->getLeftChild(), vec, ident*2);
			}
			if (wurzel->getRightChild() != NULL) {
				runTree(wurzel->getRightChild(), vec, ident*2+1);
			}
			
		};


		// ---------------------------------------------------------------------------
		// Destruktor von Heap
		~Heap() { 
			if (anzahl>0) {
				delTree(wurzel);
			};
		};
		// ---------------------------------------------------------------------------
};

// Fehlerklasse als Templateklasse
class BadAtom {	
	private:
	public:
		string getMessage(void) { return "Atom bereits vorhanden"; };
};

// Fehlerklasse
template<class T>
class NotFound {
	private:
		T atom;
	public:
		NotFound(T atom) { this->atom = atom ; };
		string getMessage(void) { return "Atom nicht gefunden"; };
		T getAtom() { return atom; };
};
